/**************************************
 * $Author: kaneda $
 * $Date: 2002/06/28 08:49:54 $
 **************************************/

/* error.c */

#include "ph_util.h"

#ifndef __PH_WIN32__
#include <syslog.h> /* for syslog() */
#else
#define LOG_INFO 	0
#define LOG_ERR 	1
#endif

int ph_daemon_proc = 0; /* set nonzero by daemon_init() */

static void ph_err_doit(int, int, const char *, va_list);




/* Nonfatal error related to a system call.
 * Print a message and return. */
void ph_err_ret(const char * fmt, ...) {
  va_list ap;
  
  va_start(ap, fmt);
  ph_err_doit(1, LOG_INFO, fmt, ap);
  va_end(ap);
  return;
}


/* Fatal error related to a system call.
 * Print a message and terminate. */
void ph_err_sys(const char * fmt, ...) {
  va_list ap;

  va_start(ap, fmt);
  ph_err_doit(1, LOG_ERR, fmt, ap);
  va_end(ap);
  exit(1);
}


/* Fatal error related to a system call.
 * Print a message, dump core, and terminate. */
void ph_err_dump(const char * fmt, ...) {
  va_list ap;

  va_start(ap, fmt);
  ph_err_doit(1, LOG_ERR, fmt, ap);
  va_end(ap);
  abort();	/* dump core and terminate */
  exit(1);	/* shouldn't get here */
}


/* Nonfatal error unrelated to a system call.
 * Print a message and return. */
void ph_err_msg(const char * fmt, ...) {
  va_list ap;
  
  va_start(ap, fmt);
  ph_err_doit(0, LOG_INFO, fmt, ap);
  va_end(ap);
  return;
}


/* Fatal error unrelated to a system call.
 * Print a message and terminate. */
void ph_err_quit(const char * fmt, ...) {
  va_list ap;

  va_start(ap, fmt);
  ph_err_doit(0, LOG_ERR, fmt, ap);
  va_end(ap);
  exit(1);
}


/* Print a message and return to caller.
 * Caller specifies "errnoflag" and "level". */
static void ph_err_doit(int errnoflag, int level, const char * fmt, va_list ap) {
  int	errno_save, n;
  char	buf[PH_MAX_LINE+1];

  errno_save = errno;	/* value caller might want printed */
#ifdef HAVE_VSNPRINTF
  vsnprintf(buf, PH_MAX_LINE, fmt, ap);
#else
  vsprintf(buf, fmt, ap);
#endif
  n = ph_strlen(buf);
  if (errnoflag) {
    snprintf(buf+n, PH_MAX_LINE - n, ": %s", strerror(errno_save));
  }
  strcat(buf, "\n");

  
  if (ph_daemon_proc) {
#ifndef __PH_WIN32__
    syslog(level, buf);
#else
    fprintf(stderr, "%d\n", level);
#endif
  } else {
    fflush(stdout);	/* in case stdout and stderr are the same */
    fputs(buf, stderr);
    fflush(stderr);
  }
    
}








