/**************************************
 * $Author: kaneda $
 * $Date: 2003/01/18 10:50:15 $
 **************************************/

/* list.c */

#include "ph_util.h"



static int is_empty(ph_list_t list) {
  if (list->nelm == 0) {
    assert(list->head == NULL);
    assert(list->tail == NULL);
  } else {
    assert(list->head != NULL);
    assert(list->tail != NULL);
  }
  return (list->nelm == 0);
}


static void lock_list(ph_list_t list) {
  if (list->is_sync) {
    ph_pthread_mutex_lock(&list->mp);
  }
}


static void unlock_list(ph_list_t list) {
  if (list->is_sync) {
    ph_pthread_mutex_unlock(&list->mp);
  }
}


static void wait_while_empty(ph_list_t list) {
  while ((list->is_sync) && (is_empty(list))) {
    ph_pthread_cond_wait(&list->i_cond, &list->mp);
  }
}


static void notify_insertion(ph_list_t list) {
  list->nelm++;
  if (list->is_sync) {
    ph_pthread_cond_broadcast(&list->i_cond);
  }
}


static void notify_deletion(ph_list_t list) {
  list->nelm--;
  if (list->is_sync) {
    ph_pthread_cond_broadcast(&list->d_cond);
  }
  assert(list->nelm >= 0);
}


int ph_list_is_empty(ph_list_t list) {
  int n;
  lock_list(list);
  n = is_empty(list);
  unlock_list(list);
  return n;
}


int ph_list_is_empty_opt(ph_list_t list, int is_sync) {
  int n;
  if (is_sync) { lock_list(list); }
  n = is_empty(list);
  if (is_sync) { unlock_list(list); }
  return n;
}


int ph_list_get_nelm(ph_list_t list) {
  int n;
  lock_list(list);
  n = list->nelm;
  unlock_list(list);
  return n;  
}

ph_list_t ph_list_new(int is_sync) {
  ph_list_t list;
  
  list = (ph_list_t)ph_malloc(sizeof(struct ph_list));
  
  list->head = NULL;
  list->tail = NULL;
  list->nelm = 0;

  list->is_sync = is_sync;

  ph_pthread_mutex_init(&list->mp, NULL);
  ph_pthread_cond_init(&list->i_cond, NULL);
  ph_pthread_cond_init(&list->d_cond, NULL);
  return list;
}


void ph_list_dispose(ph_list_t list, void (*func)(void *)) {
  ph_cell_t p, next;

  if (list == NULL) {
    ph_err_quit("ph_list_dispose() : NULL input");
  }
  
  
  for (p = list->head; p != NULL; p = next) {
    next = p->next;
    
    if (func != NULL) {
      (*func)(p->self);		/* dispose p->self */
    }
    free(p);
  }

  ph_pthread_mutex_destroy(&list->mp);
  ph_pthread_cond_destroy(&list->i_cond);
  ph_pthread_cond_destroy(&list->d_cond);
  ph_free(list);
}


/* add item to the front of the list */
void ph_list_insert(ph_list_t list, void * val) {
  ph_cell_t new_cell;
  
  if ((list == NULL) || (val == NULL)) {
    ph_err_quit("ph_list_insert() : NULL input");
  }  

  lock_list(list);

  new_cell = (ph_cell_t)ph_malloc(sizeof(struct ph_cell));
  new_cell->self = val;
  new_cell->next = list->head;

  list->head = new_cell;

  if (list->tail == NULL) {
    list->tail = new_cell;
  }

  notify_insertion(list);
  unlock_list(list);
}


static void ph_list_delete_sub(ph_list_t list, void * val, void (*func)(void *)) {
  ph_cell_t prev, p;
 
  if (is_empty(list)) {
    ph_err_quit("ph_list_delete_sub() : list is empty.");
  }

  prev = NULL;
  for (p = list->head; p != NULL; p = p->next) {
    if (p->self == val) {
      /* delete p*/

      if (prev == NULL) {	/* if P is in front */
	list->head = p->next;
	if (list->tail == p) {
	  list->tail = NULL;
	}
      } else {			/* if p is not in front */
	prev->next = p->next;
	if (list->tail == p) {
	  list->tail = prev;
	}
      }
      if (func != NULL) {
	(*func)(val);
      }
      ph_free(p);
      notify_deletion(list);
      return;		/* finish this function */
    } else {
      prev = p;
    }
  }
  ph_err_quit("ph_list_delete() : item is not in list.");
}

void ph_list_delete(ph_list_t list, void * val, void (*func)(void *)) {
  if ((list == NULL) || (val == NULL)) {
    ph_err_quit("ph_list_delete() : NULL input");
  }  

  lock_list(list);
  ph_list_delete_sub(list, val, func);
  unlock_list(list);
}


void ph_list_delete_opt(ph_list_t list, void * val, void (*func)(void *), int is_sync) {
  if ((list == NULL) || (val == NULL)) {
    ph_err_quit("ph_list_delete_opt() : NULL input");
  }
  
  if (is_sync) { ph_pthread_mutex_lock(&list->mp); }
  ph_list_delete_sub(list, val, func);
  if (is_sync) { ph_pthread_mutex_unlock(&list->mp); }
}


void ph_list_enqueue(ph_list_t list, void * val) {
  ph_cell_t new_cell;

  if ((list == NULL) || (val == NULL)) {
    ph_err_quit("ph_list_enqueue() : NULL input");
  }

  lock_list(list);

  new_cell = (ph_cell_t)ph_malloc(sizeof(struct ph_cell));
  new_cell->self = val;
  new_cell->next = NULL;
      
  if (list->head == NULL) {
    list->head = new_cell;
  } else {
    assert((list->tail != NULL) && (list->tail->next == NULL));
    list->tail->next = new_cell;
  }

  list->tail = new_cell;
  notify_insertion(list);
  unlock_list(list);
}


static void * ph_list_dequeue_sub(ph_list_t list) {
  void * val;
  ph_cell_t next;

  wait_while_empty(list);
 
  val = list->head->self;
  next = list->head->next;

  ph_free(list->head);
  list->head = next;
  
  if (list->head == NULL) {
    list->tail = NULL;
  }

  notify_deletion(list);  

  return val;
}


void * ph_list_dequeue(ph_list_t list) {
  void * val;

  if (list == NULL) {
    ph_err_quit("ph_list_dequeue() : NULL input");
  }
  
  lock_list(list);
  val = ph_list_dequeue_sub(list);
  unlock_list(list);

  return val;
}


static void * ph_list_dequeue_opt_sub_with_func(ph_list_t list, int is_block, int (*func)(void *, void *, void *), void * arg1, void * arg2) {
  void * val = NULL;

  while (val == NULL) {
    ph_cell_t prev, p;

    prev = NULL;
    for (p = list->head; p != NULL; p = p->next) {
      if ((*func)(p->self, arg1, arg2)) {
	/* dequeue p->self */
	
	if (prev == NULL) {		/* if P is in front */
	  list->head = p->next;
	  if (list->tail == p) {
	    list->tail = NULL;
	  }
	} else {			/* if p is not in front */
	  prev->next = p->next;
	  if (list->tail == p) {
	    list->tail = prev;
	  }
	}
	val = p->self;
	ph_free(p);
	notify_deletion(list);
	break;
      } else {
	prev = p;
      }
    }

    if (val == NULL) {
      if (is_block) {
	ph_pthread_cond_wait(&list->i_cond, &list->mp);
      } else {
	break;
      }
    }
  }

  return val;
}

void * ph_list_dequeue_opt(ph_list_t list, int is_block, int (*func)(void *, void *, void *), void * arg1, void * arg2) {
  void * val;

  if (list == NULL) {
    ph_err_quit("ph_list_dequeue() : NULL input");
  }
  
  lock_list(list);
  
  if ((is_block == 0) && (is_empty(list))) {
    unlock_list(list);
    return NULL;
  }

  val = (func == NULL) ? ph_list_dequeue_sub(list) : ph_list_dequeue_opt_sub_with_func(list, is_block, func, arg1, arg2);

  unlock_list(list);

  return val;
}


/* return cell which satisfies (*FUNC)(ITEM, ARG) == 1 */
void * ph_list_find(ph_list_t list, int (*func)(void *, void *), void * arg) {
  ph_cell_t p;
  void * val = NULL;

  if ((list == NULL) || (func == NULL)) {
    ph_err_quit("ph_list_find() : NULL input");
  } 

  lock_list(list);
  for (p = list->head; p != NULL; p = p->next) {
    if ((*func)(p->self, arg)) {
      val = p->self;
    }
  }
  unlock_list(list);
  return val;
}


/* map FUNC to all the items in LIST */
void ph_list_map(ph_list_t list, void (*func)(void *, void *), void * arg) {
  ph_cell_t p;

  if ((list == NULL) || (func == NULL)) {
    ph_err_quit("ph_list_map() : NULL input");
  } 

  lock_list(list);
  for (p = list->head; p != NULL; p = p->next) {
    (*func)(p->self, arg);
  }
  unlock_list(list);
}



void ph_list_sort(ph_list_t list, int (*func)(void *, void *)) {
  ph_cell_t p, q;
  
  if ((list == NULL) || (func == NULL)) {
    ph_err_quit("ph_list_sort() : NULL input");
  } 

  lock_list(list);
  for (p = list->head; p != NULL; p = p->next) {
    for (q = p; q != NULL; q = q->next) {
      if ((*func)(p->self, q->self)) {
	/* swap */
	void * tmp = p->self;
	p->self = q->self;
	q->self = tmp;
      }
    }
  }
  unlock_list(list);
}


void ph_list_swap(ph_list_t p,  ph_list_t q, int is_sync) {
  ph_cell_t tmp;
  int n;

  if ((p == NULL) || (q == NULL)) {
    ph_err_quit("ph_list_swap() : NULL input");
  }

  if (is_sync) { ph_pthread_mutex_lock(&p->mp); }
  if (is_sync) { ph_pthread_mutex_lock(&q->mp); }
  
  tmp = p->head;
  p->head = q->head;
  q->head = tmp;

  tmp = p->tail;
  p->tail = q->tail;
  q->tail = tmp;

  n = p->nelm;
  p->nelm = q->nelm;
  q->nelm = n;

  if (is_sync) { ph_pthread_mutex_unlock(&p->mp); }
  if (is_sync) { ph_pthread_mutex_unlock(&q->mp); }
}








