/**************************************
 * $Author: kaneda $
 * $Date: 2002/12/20 10:14:41 $
 **************************************/

/* list.c */

#include "util.h"



static int is_empty(list_t list) {
  if (list->nelm == 0) {
    assert(list->head == NULL);
    assert(list->tail == NULL);
  } else {
    assert(list->head != NULL);
    assert(list->tail != NULL);
  }
  return (list->nelm == 0);
}


static void lock_list(list_t list) {
  if (list->is_sync) {
    Pthread_mutex_lock(&list->mp);
  }
}


static void unlock_list(list_t list) {
  if (list->is_sync) {
    Pthread_mutex_unlock(&list->mp);
  }
}


static void wait_while_empty(list_t list) {
  while ((list->is_sync) && (is_empty(list))) {
    Pthread_cond_wait(&list->i_cond, &list->mp);
  }
}


static void notify_insertion(list_t list) {
  list->nelm++;
  if (list->is_sync) {
    Pthread_cond_broadcast(&list->i_cond);
  }
}


static void notify_deletion(list_t list) {
  list->nelm--;
  if (list->is_sync) {
    Pthread_cond_broadcast(&list->d_cond);
  }
  assert(list->nelm >= 0);
}


int list_is_empty(list_t list) {
  int n;
  lock_list(list);
  n = is_empty(list);
  unlock_list(list);
  return n;
}


int list_is_empty_opt(list_t list, int is_sync) {
  int n;
  if (is_sync) { lock_list(list); }
  n = is_empty(list);
  if (is_sync) { unlock_list(list); }
  return n;
}


int list_get_nelm(list_t list) {
  int n;
  lock_list(list);
  n = list->nelm;
  unlock_list(list);
  return n;  
}

list_t list_new(int is_sync) {
  list_t list;
  
  list = (list_t)Malloc(sizeof(struct list));
  
  list->head = NULL;
  list->tail = NULL;
  list->nelm = 0;

  list->is_sync = is_sync;

  Pthread_mutex_init(&list->mp, NULL);
  Pthread_cond_init(&list->i_cond, NULL);
  Pthread_cond_init(&list->d_cond, NULL);
  return list;
}


void list_dispose(list_t list, void (*func)(void *)) {
  cell_t p, next;

  if (list == NULL) {
    err_quit("list_dispose() : NULL input");
  }
  
  
  for (p = list->head; p != NULL; p = next) {
    next = p->next;
    
    if (func != NULL) {
      (*func)(p->self);		/* dispose p->self */
    }
    free(p);
  }

  Pthread_mutex_destroy(&list->mp);
  Pthread_cond_destroy(&list->i_cond);
  Pthread_cond_destroy(&list->d_cond);
  Free(list);
}


/* add item to the front of the list */
void list_insert(list_t list, void * val) {
  cell_t new_cell;
  
  if ((list == NULL) || (val == NULL)) {
    err_quit("list_insert() : NULL input");
  }  

  lock_list(list);

  new_cell = (cell_t)Malloc(sizeof(struct cell));
  new_cell->self = val;
  new_cell->next = list->head;

  list->head = new_cell;

  if (list->tail == NULL) {
    list->tail = new_cell;
  }

  notify_insertion(list);
  unlock_list(list);
}


static void list_delete_sub(list_t list, void * val, void (*func)(void *)) {
  cell_t prev, p;
 
  if (is_empty(list)) {
    err_quit("list_delete_sub() : list is empty.");
  }

  prev = NULL;
  for (p = list->head; p != NULL; p = p->next) {
    if (p->self == val) {
      /* delete p*/

      if (prev == NULL) {	/* if P is in front */
	list->head = p->next;
	if (list->tail == p) {
	  list->tail = NULL;
	}
      } else {			/* if p is not in front */
	prev->next = p->next;
	if (list->tail == p) {
	  list->tail = prev;
	}
      }
      if (func != NULL) {
	(*func)(val);
      }
      Free(p);
      notify_deletion(list);
      return;		/* finish this function */
    } else {
      prev = p;
    }
  }
  err_quit("list_delete() : item is not in list.");
}

void list_delete(list_t list, void * val, void (*func)(void *)) {
  if ((list == NULL) || (val == NULL)) {
    err_quit("list_delete() : NULL input");
  }  

  lock_list(list);
  list_delete_sub(list, val, func);
  unlock_list(list);
}


void list_delete_opt(list_t list, void * val, void (*func)(void *), int is_sync) {
  if ((list == NULL) || (val == NULL)) {
    err_quit("list_delete_opt() : NULL input");
  }
  
  if (is_sync) { Pthread_mutex_lock(&list->mp); }
  list_delete_sub(list, val, func);
  if (is_sync) { Pthread_mutex_unlock(&list->mp); }
}


void list_enqueue(list_t list, void * val) {
  cell_t new_cell;

  if ((list == NULL) || (val == NULL)) {
    err_quit("list_enqueue() : NULL input");
  }

  lock_list(list);

  new_cell = (cell_t)Malloc(sizeof(struct cell));
  new_cell->self = val;
  new_cell->next = NULL;
      
  if (list->head == NULL) {
    list->head = new_cell;
  } else {
    assert((list->tail != NULL) && (list->tail->next == NULL));
    list->tail->next = new_cell;
  }

  list->tail = new_cell;
  notify_insertion(list);
  unlock_list(list);
}


static void * list_dequeue_sub(list_t list) {
  void * val;
  cell_t next;

  wait_while_empty(list);
 
  val = list->head->self;
  next = list->head->next;

  Free(list->head);
  list->head = next;
  
  if (list->head == NULL) {
    list->tail = NULL;
  }

  notify_deletion(list);  

  return val;
}


void * list_dequeue(list_t list) {
  void * val;

  if (list == NULL) {
    err_quit("list_dequeue() : NULL input");
  }
  
  lock_list(list);
  val = list_dequeue_sub(list);
  unlock_list(list);

  return val;
}


static void * list_dequeue_opt_sub_with_func(list_t list, int is_block, int (*func)(void *, void *, void *), void * arg1, void * arg2) {
  void * val = NULL;

  while (val == NULL) {
    cell_t prev, p;

    prev = NULL;
    for (p = list->head; p != NULL; p = p->next) {
      if ((*func)(p->self, arg1, arg2)) {
	/* dequeue p->self */
	
	if (prev == NULL) {		/* if P is in front */
	  list->head = p->next;
	  if (list->tail == p) {
	    list->tail = NULL;
	  }
	} else {			/* if p is not in front */
	  prev->next = p->next;
	  if (list->tail == p) {
	    list->tail = prev;
	  }
	}
	val = p->self;
	Free(p);
	notify_deletion(list);
	break;
      } else {
	prev = p;
      }
    }

    if (val == NULL) {
      if (is_block) {
	Pthread_cond_wait(&list->i_cond, &list->mp);
      } else {
	break;
      }
    }
  }

  return val;
}

void * list_dequeue_opt(list_t list, int is_block, int (*func)(void *, void *, void *), void * arg1, void * arg2) {
  void * val;

  if (list == NULL) {
    err_quit("list_dequeue() : NULL input");
  }
  
  lock_list(list);
  
  if ((is_block == 0) && (is_empty(list))) {
    unlock_list(list);
    return NULL;
  }

  val = (func == NULL) ? list_dequeue_sub(list) : list_dequeue_opt_sub_with_func(list, is_block, func, arg1, arg2);

  unlock_list(list);

  return val;
}


/* return cell which satisfies (*FUNC)(ITEM, ARG) == 1 */
void * list_find(list_t list, int (*func)(void *, void *), void * arg) {
  cell_t p;
  void * val = NULL;

  if ((list == NULL) || (func == NULL)) {
    err_quit("list_find() : NULL input");
  } 

  lock_list(list);
  for (p = list->head; p != NULL; p = p->next) {
    if ((*func)(p->self, arg)) {
      val = p->self;
    }
  }
  unlock_list(list);
  return val;
}


/* map FUNC to all the items in LIST */
void list_map(list_t list, void (*func)(void *, void *), void * arg) {
  cell_t p;

  if ((list == NULL) || (func == NULL)) {
    err_quit("list_map() : NULL input");
  } 

  lock_list(list);
  for (p = list->head; p != NULL; p = p->next) {
    (*func)(p->self, arg);
  }
  unlock_list(list);
}



void list_sort(list_t list, int (*func)(void *, void *)) {
  cell_t p, q;
  
  if ((list == NULL) || (func == NULL)) {
    err_quit("list_sort() : NULL input");
  } 

  lock_list(list);
  for (p = list->head; p != NULL; p = p->next) {
    for (q = p; q != NULL; q = q->next) {
      if ((*func)(p->self, q->self)) {
	/* swap */
	void * tmp = p->self;
	p->self = q->self;
	q->self = tmp;
      }
    }
  }
  unlock_list(list);
}


void list_swap(list_t p,  list_t q, int is_sync) {
  cell_t tmp;
  int n;

  if ((p == NULL) || (q == NULL)) {
    err_quit("list_swap() : NULL input");
  }

  if (is_sync) { Pthread_mutex_lock(&p->mp); }
  if (is_sync) { Pthread_mutex_lock(&q->mp); }
  
  tmp = p->head;
  p->head = q->head;
  q->head = tmp;

  tmp = p->tail;
  p->tail = q->tail;
  q->tail = tmp;

  n = p->nelm;
  q->nelm = n;
  p->nelm = q->nelm;

  if (is_sync) { Pthread_mutex_unlock(&p->mp); }
  if (is_sync) { Pthread_mutex_unlock(&q->mp); }
}








