/**************************************
 * $Author: kaneda $
 * $Date: 2002/12/20 10:14:42 $
 **************************************/

/* msg.c */

#include "msg.h"



msg_t msg_new(ph_vp_t vp, task_id_t id, msg_type_t type, void * body, size_t len) {
  msg_t msg;

  msg = (msg_t)Malloc(sizeof(struct msg));

  msg->vp = vp;
  msg->id = id;
  msg->type = type;
  msg->body = body;
  msg->len = len;

  return msg;
}


void msg_dispose(msg_t msg) { 
  if (msg == NULL) {
    err_msg("msg_dispose() : NULL input");
    return;
  }

  if (msg->body != NULL) { Free(msg->body); }
  Free(msg);
}


static char * print_type(msg_type_t type) {
  switch (type) {
  case SPAWN_TASK:
    return "spawn_task";
    break;
  case WRITE_TERMINAL:
    return "write_terminal";
    break;
  case CLOSE_TERMINAL:
    return "close_terminal";
    break;
  case ASSUME_VP:
    return "assume_vp";
  case JOIN:
    return "join";
  case SET_ROUTE:
    return "set_route";
    break;
  case GET_MACHINE_LIST:
    return "get_machines";
  case SEND_MACHINE_LIST:
    return "send_machines";
  default:
    err_quit("print_type() : no matching (type=%d)", type);
    break;
  }
  return NULL;
}


static void msg_print_sub(msg_t msg, void * arg) {
  debug_print("(%p) vp=%lld, id=%lld, type=%s, len=%d\n",
	      msg,
	      msg->vp,
	      msg->id,
	      print_type(msg->type),
	      msg->len);
}


void msg_print(msg_t msg) {
  assert(msg != NULL);
  debug_print2("[msg \n");
  msg_print_sub(msg, NULL);
  debug_print2("]\n");
}


static void create_header(msg_t msg, char header[]) {
  Mzero(header, MSG_HEADER_SIZE);
  Snprintf(header, MSG_HEADER_SIZE, 
	   "VPG    TO: %lld %lld  TYPE %d   SIZE: %d ",
	   msg->vp, 
	   msg->id,
	   msg->type,
	   msg->len);
}


static msg_t analyze_header(char header[]) {
  int i;
  char s[3][128];
  char sign[128];
  ph_vp_t vp;
  task_id_t id;
  msg_type_t type;
  size_t len;

  Mzero(sign, 128);
  for (i = 0; i < 3; i++) {
    Mzero(s[i], 128);
  }

  sscanf(header, "%s %s%lld%lld %s%d %s%d",
	 sign,
	 s[0], &vp, &id,
	 s[1], (int *)&type,
	 s[2], &len);

  assert(Streq(sign, "VPG"));


  return msg_new(vp, id, type, Calloc(len, sizeof(char)), len);
}


void msg_send(msg_t msg, conn_t conn) {
  if (conn == NULL) { 
    char * s;
    
    s = Malloc(MSG_HEADER_SIZE + msg->len);
    create_header(msg, s);
    Mmove(s + MSG_HEADER_SIZE, msg->body, msg->len);        
    ph_send(msg->vp, s, MSG_HEADER_SIZE + msg->len, 0);
    Free(s);
  } else {

    switch (conn->type) {
    case SHELL: {
      char * s;
      
      s = Malloc(MSG_HEADER_SIZE + msg->len);
      create_header(msg, s);
      Mmove(s + MSG_HEADER_SIZE, msg->body, msg->len);    
      conn_send(conn, s, MSG_HEADER_SIZE + msg->len);
      Free(s);  
      break;
    }
    case TASK_IN:
      assert(msg->type == WRITE_TERMINAL);
      conn_send(conn, msg->body, msg->len);
      break;
    default:
      err_quit("msg_send() : no matching");
    }
  }
}


void msg_send_raw(msg_t msg, int fd) {
  char * s;
  ssize_t n;

  s = Malloc(MSG_HEADER_SIZE + msg->len);
  create_header(msg, s);
  Mmove(s + MSG_HEADER_SIZE, msg->body, msg->len);    
  n = send(fd, s, MSG_HEADER_SIZE + msg->len, 0);
  assert((n <= 0) || (n ==  MSG_HEADER_SIZE + msg->len)); 
  Free(s);  
}


msg_t msg_recv(conn_t conn) {
  switch (conn->type) {
  case SHELL:
  case PHOENIX: {
    char header[MSG_HEADER_SIZE];
    ssize_t n;
    msg_t msg;


    /* receiver a header in a message */
    Mzero(header, MSG_HEADER_SIZE);
    if ((n = conn_recv(conn, header, MSG_HEADER_SIZE)) <= 0) {
      return NULL;
    }
 
    msg = analyze_header(header);

    /* receiver a body in a message */
    if ((n = conn_recv(conn, msg->body, msg->len)) <= 0) {
      msg_dispose(msg);
      return NULL;
    }

    return msg;
    break;
  }

  case TASK_OUT: {
    char s[BUFSIZE];
    ssize_t n;

    Mzero(s, BUFSIZE);
    if ((n = conn_recv(conn, s, BUFSIZE)) <= 0) {
      return NULL;
    }
    return msg_new(conn->vp, conn->id, WRITE_TERMINAL, Mdup(s, n), n);
    break;
  }

  default:
    err_quit("msg_recv() : no matching");
  }

  return NULL;
}


msg_t msg_recv_raw(int fd) {
  char header[MSG_HEADER_SIZE];
  ssize_t n;
  msg_t msg;
  
  
  /* receiver a header in a message */
  Mzero(header, MSG_HEADER_SIZE);
  if ((n = recv(fd, header, MSG_HEADER_SIZE, MSG_WAITALL)) <= 0) {
    return NULL;
  }
  
  msg = analyze_header(header);
  
  /* receiver a body in a message */
  if ((n = recv(fd, msg->body, msg->len, MSG_WAITALL)) <= 0) {
    msg_dispose(msg);
    return NULL;
  }
  
  return msg;
}


DEFINE_LIST_FUNC_BODY(msg)


void msg_list_print(msg_list_t msg_list) {
  assert(msg_list != NULL);

  if (msg_list_is_empty(msg_list)) {
    debug_print2("[msg_list]\n");
  } else {
    debug_print2("[msg_list \n");
    msg_list_map(msg_list, &msg_print_sub, NULL);
    debug_print2("]\n");
  }
}


static int is_same_dest(msg_t msg, void * arg1, void * arg2) {
  ph_vp_t * vp_p;
  task_id_t * id_p;

  vp_p = (ph_vp_t * )arg1;
  id_p = (task_id_t * )arg2;
  /* message to the daemon (not to the shell) */
  return ((msg->vp == (*vp_p)) && (msg->id == (*id_p)));
}


msg_t msg_list_dequeue_same_dest(msg_list_t msg_list, ph_vp_t vp, task_id_t id) {
  return msg_list_dequeue_opt(msg_list, 0, &is_same_dest, (void *)&vp, (void *)&id);
}




